package build

import (
	"strconv"

	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/cpu"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/errors"
)

// Execute executes an operation on a register with a value operand.
func (f *Function) Execute(operation token.Token, register cpu.Register, value *expression.Expression) error {
	if value.IsLeaf() {
		return f.ExecuteLeaf(operation, register, value.Token)
	}

	var temporary cpu.Register

	if isFunctionCall(value) {
		temporary = f.cpu.Return[0]
	} else {
		found := false
		temporary, found = f.cpu.FindFree(f.cpu.General)

		if !found {
			panic("no free registers")
		}
	}

	f.cpu.Use(temporary)
	defer f.cpu.Free(temporary)
	err := f.ExpressionToRegister(value, temporary)

	if err != nil {
		return err
	}

	return f.ExecuteRegisterRegister(operation, register, temporary)
}

// ExecuteLeaf performs an operation on a register with the given leaf operand.
func (f *Function) ExecuteLeaf(operation token.Token, register cpu.Register, operand token.Token) error {
	switch operand.Kind {
	case token.Identifier:
		name := operand.Text()
		variable, exists := f.variables[name]

		if !exists {
			return errors.New(&errors.UnknownIdentifier{Name: name}, f.File, operand.Position)
		}

		defer f.useVariable(variable)
		return f.ExecuteRegisterRegister(operation, register, variable.Register)

	case token.Number:
		value := operand.Text()
		number, err := strconv.Atoi(value)

		if err != nil {
			return err
		}

		return f.ExecuteRegisterNumber(operation, register, number)
	}

	return errors.New(errors.NotImplemented, f.File, operation.Position)
}

// ExecuteRegisterNumber performs an operation on a register and a number.
func (f *Function) ExecuteRegisterNumber(operation token.Token, register cpu.Register, number int) error {
	switch operation.Text() {
	case "+", "+=":
		f.assembler.RegisterNumber(asm.ADD, register, number)

	case "-", "-=":
		f.assembler.RegisterNumber(asm.SUB, register, number)

	case "*", "*=":
		f.assembler.RegisterNumber(asm.MUL, register, number)

	case "/", "/=":
		f.assembler.RegisterNumber(asm.DIV, register, number)

	case "=":
		f.assembler.RegisterNumber(asm.MOVE, register, number)

	default:
		return errors.New(&errors.InvalidOperator{Operator: operation.Text()}, f.File, operation.Position)
	}

	return nil
}

// ExecuteRegisterRegister performs an operation on two registers.
func (f *Function) ExecuteRegisterRegister(operation token.Token, destination cpu.Register, source cpu.Register) error {
	switch operation.Text() {
	case "+", "+=":
		f.assembler.RegisterRegister(asm.ADD, destination, source)

	case "-", "-=":
		f.assembler.RegisterRegister(asm.SUB, destination, source)

	case "*", "*=":
		f.assembler.RegisterRegister(asm.MUL, destination, source)

	case "/", "/=":
		f.assembler.RegisterRegister(asm.DIV, destination, source)

	case "=":
		if destination != source {
			f.assembler.RegisterRegister(asm.MOVE, destination, source)
		}

	default:
		return errors.New(&errors.InvalidOperator{Operator: operation.Text()}, f.File, operation.Position)
	}

	return nil
}

// ExpressionToRegister moves the result of an expression into the given register.
func (f *Function) ExpressionToRegister(root *expression.Expression, register cpu.Register) error {
	operation := root.Token

	if root.IsLeaf() {
		return f.TokenToRegister(operation, register)
	}

	if isFunctionCall(root) {
		err := f.CompileFunctionCall(root)

		if err != nil {
			return err
		}

		if register != f.cpu.Return[0] {
			f.assembler.RegisterRegister(asm.MOVE, register, f.cpu.Return[0])
		}

		return nil
	}

	left := root.Children[0]
	right := root.Children[1]

	err := f.ExpressionToRegister(left, register)

	if err != nil {
		return err
	}

	f.SaveRegister(register)
	return f.Execute(operation, register, right)
}

// ExpressionsToRegisters moves multiple expressions into the specified registers.
func (f *Function) ExpressionsToRegisters(expressions []*expression.Expression, registers []cpu.Register) error {
	for i := len(expressions) - 1; i >= 0; i-- {
		expression := expressions[i]
		register := registers[i]

		err := f.ExpressionToRegister(expression, register)

		if err != nil {
			return err
		}
	}

	return nil
}

// TokenToRegister moves a token into a register.
// It only works with identifiers, numbers and strings.
func (f *Function) TokenToRegister(t token.Token, register cpu.Register) error {
	switch t.Kind {
	case token.Identifier:
		name := t.Text()
		variable, exists := f.variables[name]

		if !exists {
			return errors.New(&errors.UnknownIdentifier{Name: name}, f.File, t.Position)
		}

		if register != variable.Register {
			f.assembler.RegisterRegister(asm.MOVE, register, variable.Register)
		}

		f.useVariable(variable)
		return nil

	case token.Number:
		value := t.Text()
		n, err := strconv.Atoi(value)

		if err != nil {
			return err
		}

		f.assembler.RegisterNumber(asm.MOVE, register, n)
		return nil

	case token.String:
		return errors.New(errors.NotImplemented, f.File, t.Position)

	default:
		return errors.New(errors.InvalidExpression, f.File, t.Position)
	}
}
